/*****************************************************
GNU MP interface for Objective CAML

v0.13
David.Monniaux@ens-lyon.fr
*****************************************************/

#include <gmp.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <stdio.h>

#define WORDLENGTH(x) ((sizeof(x)+sizeof(header_t)-1)/sizeof(header_t))

/******************************* MPZ *******************************/
#define MPZ_VAL(x) (((struct boxed_mpz*)(x))->val)

struct boxed_mpz
{
  void (*finalize)(struct boxed_mpz*);
  mpz_t val;
};

void mpz_finalize(struct boxed_mpz *ptr)
{
  mpz_clear(ptr->val);
}

/* Alloc */
struct boxed_mpz* alloc_mpz()
{
  struct boxed_mpz* r=(struct boxed_mpz*)
    alloc_shr(WORDLENGTH(struct boxed_mpz), Final_tag);
  r->finalize=mpz_finalize;
  return r;
}

/* Initialization functions */
value ml_mpz_from_int(value v)
{
  struct boxed_mpz *ret = alloc_mpz();
  mpz_init_set_si(ret->val, Long_val(v));
  return (value) ret;
}

value ml_mpz2_from_int(value r, value v)
{
  mpz_set_si(MPZ_VAL(r), Long_val(v));
  return r;
}

value ml_mpz_from_string(value str, value base)
{
  struct boxed_mpz *ret = alloc_mpz();
  if (mpz_init_set_str(ret->val, String_val(str), Int_val(base)))
    invalid_argument("Gmp.Z.from_string");
  return (value) ret;
}

value ml_mpz2_from_string(value r, value str, value base)
{
  if (mpz_set_str(MPZ_VAL(r), String_val(str), Int_val(base)))
    invalid_argument("Gmp.Z2.from_string");
  return r;
}

value ml_mpz_from_float(value f)
{
  struct boxed_mpz *ret = alloc_mpz();
  mpz_init_set_d(ret->val, Double_val(f));
  return (value) ret;
}

value ml_mpz2_from_float(value r, value f)
{
  mpz_set_d(MPZ_VAL(r), Double_val(f));
  return r;
}

/* Conversion to other types*/
value ml_int_from_mpz(value v)
{
  return Val_long(mpz_get_si(MPZ_VAL(v)));
}

value ml_float_from_mpz(value v)
{
  value ret=alloc(1, Double_tag);
  Store_double_val(ret, mpz_get_d(MPZ_VAL(v)));
  return ret;
}

value ml_string_from_mpz(value v, value base)
{
  char *s=mpz_get_str(NULL, Int_val(base), MPZ_VAL(v));
  value ret=copy_string(s);
  free(s);
  return ret;
}

/* Alloc and init */
struct boxed_mpz* alloc_init_mpz()
{
  struct boxed_mpz* r=(struct boxed_mpz*)
    alloc_shr(WORDLENGTH(struct boxed_mpz), Final_tag);
  r->finalize=mpz_finalize;
  mpz_init(r->val);
  return r;
}

value ml_mpz_copy(value x)
{
  struct boxed_mpz* r=(struct boxed_mpz*)
    alloc_shr(WORDLENGTH(struct boxed_mpz), Final_tag);
  r->finalize=mpz_finalize;
  mpz_init_set(r->val, MPZ_VAL(x));
  return (value) r;
}

value ml_mpz2_copy(value r, value x)
{
  mpz_set(MPZ_VAL(r), MPZ_VAL(x));
  return r;
}

/* Binary ops (except division) */
#define MPZ_l2(func) \
value ml_mpz_##func (value x, value y) \
{ \
  struct boxed_mpz* r=alloc_init_mpz(); \
  mpz_##func (r->val, MPZ_VAL(x), MPZ_VAL(y)); \
  return (value) r; \
} \
\
value ml_mpz2_##func (value r, value x, value y) \
{ \
  mpz_##func (MPZ_VAL(r), MPZ_VAL(x), MPZ_VAL(y)); \
  return r; \
}

MPZ_l2(and)
MPZ_l2(ior)

#define MPZ_u2(func) \
value ml_mpz_##func (value x, value y) \
{ \
  struct boxed_mpz* r=alloc_init_mpz(); \
  mpz_##func (r->val, MPZ_VAL(x), Long_val(y)); \
  return (value) r; \
} \
\
value ml_mpz2_##func (value r, value x, value y) \
{ \
  mpz_##func (MPZ_VAL(r), MPZ_VAL(x), Long_val(y)); \
  return r; \
}

#define MPZ_2(func) \
MPZ_l2(func) \
MPZ_u2(func##_ui)

MPZ_2(add)
MPZ_2(sub)
MPZ_2(mul)

/* Exponentiation */
MPZ_u2(pow_ui)

value ml_mpz_ui_pow_ui(value x, value y)
{
  struct boxed_mpz* r=alloc_init_mpz();
  mpz_ui_pow_ui(r->val, Long_val(x), Long_val(y));
  return (value) r;
}

value ml_mpz2_ui_pow_ui(value r, value x, value y)
{
  mpz_ui_pow_ui(MPZ_VAL(r), Long_val(x), Long_val(y));
  return r;
}

value ml_mpz_powm(value base, value exp, value mod)
{
  struct boxed_mpz* r=alloc_init_mpz();
  mpz_powm(r->val, MPZ_VAL(base), MPZ_VAL(exp), MPZ_VAL(mod));
  return (value) r;
}

value ml_mpz2_powm(value r, value base, value exp, value mod)
{
  mpz_powm(MPZ_VAL(r), MPZ_VAL(base), MPZ_VAL(exp), MPZ_VAL(mod));
  return r;
}

value ml_mpz_powm_ui(value base, value exp, value mod)
{
  struct boxed_mpz* r=alloc_init_mpz();
  mpz_powm_ui(r->val, MPZ_VAL(base), Long_val(exp), MPZ_VAL(mod));
  return (value) r;
}

value ml_mpz2_powm_ui(value r, value base, value exp, value mod)
{
  mpz_powm_ui(MPZ_VAL(r), MPZ_VAL(base), Long_val(exp), MPZ_VAL(mod));
  return r;
}

value ml_mpz2_setbit(value r, value exp)
{
  mpz_setbit(MPZ_VAL(r), Long_val(exp));
  return r;
}

value ml_mpz2_clrbit(value r, value exp)
{
  mpz_clrbit(MPZ_VAL(r), Long_val(exp));
  return r;
}

/* Unary ops */
#define MPZ_1(func) \
value ml_mpz_##func (value x) \
{ \
  struct boxed_mpz* r=alloc_init_mpz(); \
  mpz_##func (r->val, MPZ_VAL(x)); \
  return (value) r; \
} \
\
value ml_mpz2_##func (value r, value x) \
{ \
  mpz_##func (MPZ_VAL(r), MPZ_VAL(x)); \
  return r; \
}

MPZ_1(neg)
MPZ_1(abs)
MPZ_1(com)

/* Square roots */
MPZ_1(sqrt)

value ml_mpz_sqrtrem(value x)
{
  struct boxed_mpz* r1=alloc_init_mpz(), *r2=alloc_init_mpz();
  value ret;
  mpz_sqrtrem(r1->val, r2->val, MPZ_VAL(x));
  ret = alloc(2, 0);
  Field(ret, 0) = (value) r1;
  Field(ret, 1) = (value) r2;
  return ret;
}

value ml_mpz_perfect_square_p(value x)
{
  return mpz_perfect_square_p(MPZ_VAL(x)) ? Val_true : Val_false;
}

/* Misc */
value ml_mpz_mul2exp(value x, value shift)
{
  struct boxed_mpz* r=alloc_init_mpz();
  mpz_mul_2exp(r->val, MPZ_VAL(x), Long_val(shift));
  return (value) r;
}

value ml_mpz2_mul2exp(value r, value x, value shift)
{
  mpz_mul_2exp(MPZ_VAL(r), MPZ_VAL(x), Long_val(shift));
  return r;
}

#define MPZ_u1(func) \
value ml_mpz_##func (value x) \
{ \
  struct boxed_mpz* r=alloc_init_mpz(); \
  mpz_##func (r->val, Long_val(x)); \
  return (value) r; \
} \
\
value ml_mpz2_##func (value r, value x) \
{ \
  mpz_##func (MPZ_VAL(r), Long_val(x)); \
  return r; \
}

MPZ_u1(fac_ui)
MPZ_u1(random)
MPZ_u1(random2)

/* Division */
#define MPZ_l2_p(func) \
value ml_mpz_##func (value x, value y) \
{ \
  struct boxed_mpz* r1=alloc_init_mpz(), *r2=alloc_init_mpz(); \
  value ret; \
  mpz_##func(r1->val, r2->val, MPZ_VAL(x), MPZ_VAL(y)); \
  ret = alloc(2, 0); \
  Field(ret, 0) = (value) r1; \
  Field(ret, 1) = (value) r2; \
  return ret; \
}

#define MPZ_2_p(func) \
MPZ_l2_p(func) \
value ml_mpz_##func##_ui (value x, value y) \
{ \
  struct boxed_mpz* r1=alloc_init_mpz(), *r2=alloc_init_mpz(); \
  value ret; \
  mpz_##func##_ui (r1->val, r2->val, MPZ_VAL(x), Long_val(y)); \
  ret = alloc(2, 0); \
  Field(ret, 0) = (value) r1; \
  Field(ret, 1) = (value) r2; \
  return ret; \
}

#define MPZ_div(letter) \
  MPZ_2(letter##div_q) \
  MPZ_2(letter##div_r) \
  MPZ_2_p(letter##div_qr)

#define MPZ_div2exp(letter) \
  MPZ_u2(letter##div_q_2exp) \
  MPZ_u2(letter##div_r_2exp) \

MPZ_div(t)
MPZ_div(f)
MPZ_div(c)

MPZ_div2exp(t)
MPZ_div2exp(f)

MPZ_2(mod)
MPZ_l2(divexact)

/* Compare */
#define MPZ_il2(func) \
value ml_mpz_##func (value x, value y) \
{ \
  return Val_int(mpz_##func (MPZ_VAL(x), \
    MPZ_VAL(y))); \
}

#define MPZ_iu2(func) \
value ml_mpz_##func (value x, value y) \
{ \
  return Val_int(mpz_##func (MPZ_VAL(x), Long_val(y))); \
}

#define MPZ_i2(func) \
MPZ_il2(func) \
MPZ_iu2(func##_ui) 

MPZ_il2(cmp)
MPZ_iu2(cmp_si)

#define MPZ_i1(func) \
value ml_mpz_##func (value x) \
{ \
  return Val_int(mpz_##func (MPZ_VAL(x))); \
}

MPZ_i1(sgn)

MPZ_i1(popcount)
MPZ_il2(hamdist)

MPZ_iu2(scan0)
MPZ_iu2(scan1)

/* Number theoretic */
MPZ_2(gcd)

value ml_mpz_probab_prime_p(value x, value y)
{
  return mpz_probab_prime_p(MPZ_VAL(x), Int_val(y)) ? Val_true : Val_false;
}

value ml_mpz_gcdext(value x, value y)
{
  value r;
  struct boxed_mpz *r1=alloc_init_mpz(), *r2=alloc_init_mpz(),
                   *r3=alloc_init_mpz();
  mpz_gcdext(r1->val, r2->val, r3->val, MPZ_VAL(x), MPZ_VAL(y));
  r=alloc(3, 0);
  Field(r,0) = (value) r1;
  Field(r,1) = (value) r2;
  Field(r,2) = (value) r3;
  return r;
}

MPZ_il2(jacobi)
MPZ_il2(legendre)

value ml_mpz_invert(value x, value mod)
{
  struct boxed_mpz *r=alloc_init_mpz();
  if (! mpz_invert(r->val, MPZ_VAL(x), MPZ_VAL(mod)))
    failwith("Gmp.Z.invert");
  return (value) r;
}

value ml_mpz2_invert(value r, value x, value mod)
{
  if (! mpz_invert(MPZ_VAL(r), MPZ_VAL(x), MPZ_VAL(mod)))
    failwith("Gmp.Z2.invert");
  return r;
}

/******************************* MPQ *******************************/
#define MPQ_VAL(x) (((struct boxed_mpq*)(x))->val)

struct boxed_mpq
{
  void (*finalize)(struct boxed_mpq*);
  mpq_t val;
};

void mpq_finalize(struct boxed_mpq *ptr)
{
  mpq_clear(ptr->val);
}

struct boxed_mpq* alloc_init_mpq()
{
  struct boxed_mpq* r=(struct boxed_mpq*)
    alloc_shr(WORDLENGTH(struct boxed_mpq), Final_tag);
  r->finalize=mpq_finalize;
  mpq_init(r->val);
  return r;
}

value ml_mpq_from_ints(value x, value y)
{
  struct boxed_mpq *r= alloc_init_mpq();
  mpq_set_si(r->val, Long_val(x), Long_val(y));
  return (value) r;
}

value ml_mpq2_from_ints(value r, value x, value y)
{
  mpq_set_si(MPQ_VAL(r), Long_val(x), Long_val(y));
  return r;
}

value ml_mpq_from_z(value x)
{
  struct boxed_mpq *r= alloc_init_mpq();
  mpq_set_z(r->val, MPZ_VAL(x));
  mpq_canonicalize(r->val);
  return (value) r;
}

value ml_mpq2_from_z(value r, value x)
{
  mpq_set_z(MPQ_VAL(r), MPZ_VAL(x));
  mpq_canonicalize(MPQ_VAL(r));
  return r;
}

value ml_float_from_mpq(value v)
{
  value ret=alloc(1, Double_tag);
  Store_double_val(ret, mpq_get_d(MPQ_VAL(v)));
  return ret;
}

value ml_mpq_copy(value x)
{
  struct boxed_mpq* r=(struct boxed_mpq*)
    alloc_shr(WORDLENGTH(struct boxed_mpq), Final_tag);
  r->finalize=mpq_finalize;
  mpq_init(r->val);
  mpq_set(r->val, MPQ_VAL(x));
  return (value) r;
}

value ml_mpq2_copy(value r, value x)
{
  mpq_set(MPQ_VAL(r), MPQ_VAL(x));
  return r;
}

#define MPQ_2(func) \
value ml_mpq_##func (value x, value y) \
{ \
  struct boxed_mpq *r= alloc_init_mpq(); \
  mpq_##func (r->val, MPQ_VAL(x), \
    MPQ_VAL(y)); \
  return (value) r; \
} \
\
value ml_mpq2_##func (value r, value x, value y) \
{ \
  mpq_##func (MPQ_VAL(r), MPQ_VAL(x), MPQ_VAL(y)); \
  return r; \
}

MPQ_2(add)
MPQ_2(sub)
MPQ_2(mul)
MPQ_2(div)

#define MPQ_1(func) \
value ml_mpq_##func (value x) \
{ \
  struct boxed_mpq *r= alloc_init_mpq(); \
  mpq_##func (r->val, MPQ_VAL(x) ); \
  return (value) r; \
} \
\
value ml_mpq2_##func (value r, value x) \
{ \
  mpq_##func (MPQ_VAL(r), MPQ_VAL(x) ); \
  return r; \
}

MPQ_1(neg)
MPQ_1(inv)

#define MPQZ_1(func) \
value ml_mpq_##func (value x) \
{ \
  struct boxed_mpz *r= alloc_init_mpz(); \
  mpq_##func (r->val, MPQ_VAL(x) ); \
  return (value) r; \
} \
\
value ml_mpq2_##func (value r, value x) \
{ \
  mpq_##func (MPZ_VAL(r), MPQ_VAL(x) ); \
  return r; \
}

MPQZ_1(get_num)
MPQZ_1(get_den)

#define MPQ_i2(func) \
value ml_mpq_##func (value x, value y) \
{ \
  return Val_int(mpq_##func (MPQ_VAL(x), MPQ_VAL(y)) ); \
}

MPQ_i2(cmp)

#define MPQ_i1(func) \
value ml_mpq_##func (value x) \
{ \
  return Val_int(mpq_##func (MPQ_VAL(x) )); \
}

MPQ_i1(sgn)

value ml_mpq_equal (value x, value y)
{
  return mpq_equal (MPQ_VAL(x), MPQ_VAL(y))  ? Val_true : Val_false;
}

/******************************* MPF *******************************/
#define MPF_VAL(x) (((struct boxed_mpf*)(x))->val)

struct boxed_mpf
{
  void (*finalize)(struct boxed_mpf*);
  mpf_t val;
};

void mpf_finalize(struct boxed_mpf *ptr)
{
  mpf_clear(ptr->val);
}

struct boxed_mpf* alloc_mpf()
{
  struct boxed_mpf* r=(struct boxed_mpf*)
    alloc_shr(WORDLENGTH(struct boxed_mpf), Final_tag);
  r->finalize=mpf_finalize;
  return r;
}

struct boxed_mpf* alloc_init_mpf()
{
  struct boxed_mpf* r=(struct boxed_mpf*)
    alloc_shr(WORDLENGTH(struct boxed_mpf), Final_tag);
  r->finalize=mpf_finalize;
  mpf_init(r->val);
  return r;
}

struct boxed_mpf* alloc_init_mpf2(value prec)
{
  struct boxed_mpf* r=(struct boxed_mpf*)
    alloc_shr(WORDLENGTH(struct boxed_mpf), Final_tag);
  r->finalize=mpf_finalize;
  mpf_init2(r->val, Long_val(prec));
  return r;
}

value ml_mpf_set_default_prec(value prec)
{
  mpf_set_default_prec(Long_val(prec));
  return Val_unit;
}

value ml_mpf_from_int(value v)
{
  struct boxed_mpf *ret = alloc_mpf();
  mpf_init_set_si(ret->val, Long_val(v));
  return (value) ret;
}

value ml_mpf_from_int_prec(value v, value p)
{
  struct boxed_mpf *ret = alloc_init_mpf2(p);
  mpf_set_si(ret->val, Long_val(v));
  return (value) ret;
}

value ml_mpf2_from_int(value r, value v)
{
  mpf_set_si(MPF_VAL(r), Long_val(v));
  return r;
}

value ml_mpf_from_string(value str, value base)
{
  struct boxed_mpf *ret = alloc_mpf();
  if (mpf_init_set_str(ret->val, String_val(str), Int_val(base)))
    invalid_argument("Gmp.F.from_string");
  return (value) ret;
}

value ml_mpf_from_string_prec(value str, value base, value p)
{
  struct boxed_mpf *ret = alloc_init_mpf2(p);
  if (mpf_set_str(ret->val, String_val(str), Int_val(base)))
    invalid_argument("Gmp.F.from_string_prec");
  return (value) ret;
}

value ml_mpf2_from_string(value r, value str, value base)
{
  if (mpf_set_str(MPF_VAL(r), String_val(str), Int_val(base)))
    invalid_argument("Gmp.F2.from_string");
  return r;
}

value ml_mpf_from_float(value f)
{
  struct boxed_mpf *ret = alloc_mpf();
  mpf_init_set_d(ret->val, Double_val(f));
  return (value) ret;
}

value ml_mpf_from_float_prec(value f, value p)
{
  struct boxed_mpf *ret = alloc_init_mpf2(p);
  mpf_set_d(ret->val, Double_val(f));
  return (value) ret;
}

value ml_mpf2_from_float(value f)
{
  struct boxed_mpf *ret = alloc_mpf();
  mpf_init_set_d(ret->val, Double_val(f));
  return (value) ret;
}

value ml_mpf_copy(value x)
{
  /*  struct boxed_mpf* r=(struct boxed_mpf*)
    alloc_shr(WORDLENGTH(struct boxed_mpf), Final_tag);
  r->finalize=mpf_finalize;
  mpf_init(r->val); */

  struct boxed_mpf* r=alloc_init_mpf();
  mpf_set(r->val, MPF_VAL(x));
  return (value) r;
}

value ml_mpf_copy_prec(value x, value p)
{
  struct boxed_mpf* r=alloc_init_mpf2(p);
  mpf_set(r->val, MPF_VAL(x));
  return (value) r;
}

value ml_mpf2_copy(value r, value x)
{
  mpf_set(MPF_VAL(r), MPF_VAL(x));
  return r;
}

/* Conversion to other types*/
value ml_float_from_mpf(value v)
{
  value ret=alloc(1, Double_tag);
  Store_double_val(ret, mpf_get_d(MPF_VAL(v)));
  return ret;
}

value ml_string_exp_from_mpf(value v, value base, value digits)
{
  mp_exp_t expo;
  char *s=mpf_get_str(NULL, &expo, Int_val(base), Long_val(digits), MPF_VAL(v));
  value copys=copy_string(s);
  value ret;
  free(s);
  ret=alloc(2, 0);
  Field(ret, 0) = copys;
  Field(ret, 1) = Val_long(expo);
  return ret;
}

/* Operations */
#define MPF_l2(func) \
value ml_mpf_##func (value x, value y) \
{ \
  struct boxed_mpf* r=alloc_init_mpf(); \
  mpf_##func (r->val, MPF_VAL(x), MPF_VAL(y)); \
  return (value) r; \
} \
\
value ml_mpf2_##func (value r, value x, value y) \
{ \
  mpf_##func (MPF_VAL(r), MPF_VAL(x), MPF_VAL(y)); \
  return r; \
}

#define MPF_u2(func) \
value ml_mpf_##func (value x, value y) \
{ \
  struct boxed_mpf* r=alloc_init_mpf(); \
  mpf_##func (r->val, MPF_VAL(x), Long_val(y)); \
  return (value) r; \
} \
\
value ml_mpf2_##func (value r, value x, value y) \
{ \
  mpf_##func (MPF_VAL(r), MPF_VAL(x), Long_val(y)); \
  return r; \
}

#define MPF_v2(func) \
value ml_mpf_##func (value x, value y) \
{ \
  struct boxed_mpf* r=alloc_init_mpf(); \
  mpf_##func (r->val, Long_val(x), MPF_VAL(y)); \
  return (value) r; \
} \
\
value ml_mpf2_##func (value r, value x, value y) \
{ \
  mpf_##func (MPF_VAL(r), Long_val(x), MPF_VAL(y)); \
  return r; \
}

#define MPF_2(func)\
MPF_l2(func) \
MPF_u2(func##_ui)

#define MPF_d2(func)\
MPF_2(func) \
MPF_v2(ui_##func)

MPF_2(add)
MPF_2(mul)
MPF_d2(sub)
MPF_d2(div)

#define MPF_l1(func) \
value ml_mpf_##func (value x) \
{ \
  struct boxed_mpf* r=alloc_init_mpf(); \
  mpf_##func (r->val, MPF_VAL(x)); \
  return (value) r; \
} \
\
value ml_mpf2_##func (value r, value x) \
{ \
  mpf_##func (MPF_VAL(r), MPF_VAL(x)); \
  return r; \
}

#define MPF_u1(func) \
value ml_mpf_##func (value x) \
{ \
  struct boxed_mpf* r=alloc_init_mpf(); \
  mpf_##func (r->val, Long_val(x)); \
  return (value) r; \
} \
\
value ml_mpf2_##func (value r, value x) \
{ \
  mpf_##func (MPF_VAL(r), Long_val(x)); \
  return r; \
}

#define MPF_1(func) \
MPF_l1(func) \
MPF_u1(func##_ui)

MPF_1(sqrt)

MPF_l1(neg)
MPF_l1(abs)

MPF_u2(mul_2exp)
MPF_u2(div_2exp)

#define MPF_il2(func) \
value ml_mpf_##func (value x, value y) \
{ \
  return Val_int(mpf_##func (MPF_VAL(x), MPF_VAL(y))); \
}

#define MPF_iu2(func) \
value ml_mpf_##func (value x, value y) \
{ \
  return Val_int(mpf_##func (MPF_VAL(x), Long_val(y))); \
}

MPF_il2(cmp)
MPF_iu2(cmp_si)

MPF_l2(reldiff)

#define MPF_il1(func) \
value ml_mpf_##func (value x) \
{ \
  return Val_int(mpf_##func (MPF_VAL(x))); \
}

MPF_il1(sgn)

value ml_mpf_eq(value x, value y, value prec)
{
  return Val_int(mpf_eq(MPF_VAL(x), MPF_VAL(y), Int_val(prec)));
}

value ml_mpf_random2 (value max_size, value max_exp)
{
  struct boxed_mpf* r=alloc_init_mpf();
  mpf_random2 (r->val, Long_val(max_size), Long_val(max_exp));
  return (value) r;
}

value ml_mpf2_random2 (value r, value max_size, value max_exp)
{
  mpf_random2 (MPF_VAL(r), Long_val(max_size), Long_val(max_exp));
  return r;
}

